#--------------------------------------------------------------------------
#
#   ALHM - Swarm
#
#--------------------------------------------------------------------------

from __future__ import division
from pygame import draw, Rect
from functions import *

def force(c, x):
    m = 50
    return max(-m, min(m, c - x))

def collide_x(x0, y0, x1, y1, wx, wy0, wy1):
    #print "collide_x: x0, y0 =", (x0, y0), "x1, y1 =", (x1, y1), "wx =", wx, "wy0, wy1 =", (wy0, wy1)
    dx = x1 - x0
    if dx:
        t = (wx - x0) / dx
        #print "t =", t
        if 0.0 <= t < 1.0:
            dy = y1 - y0
            wy = y0 + t * dy
            if wy >= wy0 and wy <= wy1:
                x1 = x0 + 0.75 * t * dx
                y1 = y0 + 0.75 * t * dy
    return x0, y0, x1, y1

#--------------------------------------------------------------------------

class Swarm(object):

    sounds = [get_sound("buzz-%s.ogg" % (i + 1)) for i in xrange(12)]

    initial_size = 10
#     initial_size = 1

#     centre = [300, 300]
    centre = [64, 64]

    speed = 50
#     speed = 100

    time_between_sounds = 0.5

    def __init__(self, pos):
        self.centre = list(pos)
        self.positions = [random_point(self.centre, 20) for i in xrange(self.initial_size)]
        self.dead_positions = []
#         self.velocities = [random_point((0, 0), 20) for i in xrange(self.initial_size)]
#         self.target_cc = None
        self.target_point = None
        self.velocity = (0, 0)
        self.sound_timer = 0
    
    @property
    def size(self):
        return len(self.positions)

    def draw(self, surf):
        #self.debug_draw_centre(surf)
        self.draw_positions(surf, self.positions)
        self.draw_positions(surf, self.dead_positions)
    
    def debug_draw_centre(self, surf):
        r = Rect(0, 0, 10, 10)
        r.center = self.centre
        draw.rect(surf, black, r, 1)
    
    def draw_positions(self, surf, positions):
        x0 = hornet.get_width() // 2
        y0 = hornet.get_height() // 2
        for x1, y1 in positions:
            x = int(round(x1 - x0))
            y = int(round(y1 - x0))
            if x >= 0 and y >= 0:
                surf.blit(hornet, (x, y))

    def begin_frame(self, level, dt):
        self.think(level)
        self.physics_step(level, dt)
        self.check_traps(level, dt)
        self.update_sound(dt)
    
    def update_sound(self, dt):
        if self.sound_timer > 0:
            self.sound_timer -= dt
        else:
            self.sound_timer = self.time_between_sounds
            sound = choice(self.sounds)
            sound.set_volume(0.5 * self.size / self.initial_size)
            sound.play()
    
    def think(self, level):
        if self.target_point is not None:
            if points_near(self.centre, self.target_point):
                self.target_point = None
        if self.target_point is None:
            self.select_target_point(level)
    
    def select_target_point(self, level):
        tcc = self.select_target_cc(level)
        if tcc is not None:
            self.target_point = tp = level.cell_centre(tcc)
#             print "Swarm.think: target_point =", self.target_point
            if tp is not None:
                self.velocity = scale(self.speed, unit(sub(tp, self.centre)))
        
    def select_target_cc(self, level):
        return self.head_for_player(level) or self.head_for_goal(level)
    
    def head_for_player(self, level):
        cc = level.cc_containing(self.centre)
        cell = level.get_cell(cc)
        if cell.distance_from_goal < 3:
            return None
        pcc = level.cc_containing(level.player.position)
        if pcc == cc:
            return cc
        if points_near(cc, pcc, 3):
            dirs = directions_to_cc(cc, pcc)
            for dir in dirs:
                if cell.connected[dir]:
                    return cc_in_direction(cc, dir)
    
    def head_for_goal(self, level):
        cc = level.cc_containing(self.centre)
        cell = level.get_cell(cc)
        if cell:
            best_ccs = []
            best_distance = cell.distance_from_goal
            if best_distance == 0:
                level.game_outcome = -1
                return None
            for dir in cell.connected_directions():
                cc2 = cc_in_direction(cc, dir)
                cell = level.get_cell(cc2)
                if cell:
                    cell_distance = cell.distance_from_goal
                    if cell_distance < best_distance:
                        del best_ccs[:]
                    if cell_distance <= best_distance:
                        best_ccs.append(cc2)
                        best_distance = cell_distance
            if best_ccs:
                return choice(best_ccs)
    
    def physics_step(self, level, dt):
        a1 = 2.0
        a2 = 200.0
        cx, cy = self.centre
        positions = self.positions
        cww = self.collide_with_walls
        walls = level.walls
        for p in positions:
            x0, y0 = p
            fx = force(cx, x0)
            fy = force(cy, y0)
            dx = normalvariate(0, a2)
            dy = normalvariate(0, a2)
            x1 = x0 + dt * (a1 * fx + dx)
            y1 = y0 + dt * (a1 * fy + dy)
            p[0], p[1] = cww(walls, x0, y0, x1, y1)
        scale_add_to(dt, self.velocity, self.centre)

    def collide_with_walls(self, walls, x0, y0, x1, y1):
        cx = collide_x
        for wall in walls:
            if wall.is_solid():
                (wx0, wy0), (wx1, wy1) = wall.points
                #print "wall", (wx0, wy0), (wx1, wy1)
                if wx0 == wx1:
                    #print "collide with vertical wall"
                    x0, y0, x1, y1 = cx(x0, y0, x1, y1, wx0, wy0, wy1)
                else:
                    #print "collide with horizontal wall"
                    y0, x0, y1, x1 = cx(y0, x0, y1, x1, wy0, wx0, wx1)
        return x1, y1

    def check_traps(self, level, dt):
        positions = self.positions
        for trap in level.traps:
            if trap.is_operational():
                r = trap.bounds
                i = 0
                while i < len(positions):
                    p = positions[i]
                    if r.collidepoint(p) and chance(dt * trap.catch_rate):
                        del positions[i]
                        self.dead_positions.append(p)
                    else:
                        i += 1
